# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.


from typing import Literal, overload

from boulderopalscaleupsdk.common.dtypes import ISO8601DatetimeUTCLike
from boulderopalscaleupsdk.device.controller import (
    QBLOXControllerInfo,
    QuantumMachinesControllerInfo,
)
from boulderopalscaleupsdk.device.defcal import DefCalData
from boulderopalscaleupsdk.device.processor import SuperconductingProcessor
from boulderopalscaleupsdk.experiments.classifiers import (
    Classifier,
    LinearIQClassifier,
    LinearNDClassifier,
)
from pydantic import BaseModel, Field
from pydantic.dataclasses import dataclass
from rich.console import Console

DeviceName = str


@dataclass
class Defcal:
    gate: str
    addr: str | tuple[str, ...]
    program: str

    def show(self) -> None:
        console = Console()
        console.is_jupyter = False
        console.print(self.program)


@dataclass
class DeviceData:
    qpu: SuperconductingProcessor
    controller_info: QBLOXControllerInfo | QuantumMachinesControllerInfo
    _defcals: dict[tuple[str, tuple[str, ...]], DefCalData]
    _iq_classifiers: dict[str, LinearIQClassifier]
    _leakage_classifiers: dict[tuple[str, str], LinearNDClassifier]
    _enabled_qubits: list[str]

    def get_defcal(self, gate: str, addr: str | tuple[str, ...]) -> Defcal:
        """
        Get the defcal for a specific gate and address alias.
        """
        if self._defcals == {}:
            raise ValueError("No defcal data available.")
        _addr = (addr,) if isinstance(addr, str) else tuple(i for i in sorted(addr))
        defcal = self._defcals.get((gate, _addr))
        if defcal is None:
            raise KeyError(f"No defcal data found for gate '{gate}' and address '{_addr}'.")
        return Defcal(gate=gate, addr=_addr, program=defcal.program)

    @overload
    def get_classifier(
        self,
        dtype: Literal["linear_iq"],
        addr: str | tuple[str],
    ) -> LinearIQClassifier: ...

    @overload
    def get_classifier(
        self,
        dtype: Literal["linear_nd"],
        addr: tuple[str, str],
    ) -> LinearNDClassifier: ...

    def get_classifier(
        self,
        dtype: Literal["linear_iq", "linear_nd"],
        addr: str | tuple[str, ...],
    ) -> Classifier:
        """
        Get a classifier for an address alias of a specific type.
        """
        _addr = (addr,) if isinstance(addr, str) else tuple(i for i in sorted(addr))
        classifier: Classifier | None
        match dtype:
            case "linear_iq":
                if len(_addr) != 1:
                    raise ValueError(f"Invalid address '{addr}' for '{dtype}' classifier.")
                classifier = self._iq_classifiers.get(_addr[0])
            case "linear_nd":
                if len(_addr) != 2:
                    raise ValueError(f"Invalid address '{addr}' for '{dtype}' classifier.")
                classifier = self._leakage_classifiers.get(_addr)
            case _:
                raise ValueError(f"Unsupported classifier dtype '{dtype}'.")

        if classifier is None:
            raise KeyError(f"No classifier found of type '{dtype}' and address '{addr}'.")
        return classifier

    def get_enabled_qubits(self) -> list[str] | Literal["all"]:
        match self._enabled_qubits:
            case []:
                return "all"
            case _:
                return self._enabled_qubits


class DeviceSummary(BaseModel):
    id: str
    organization_id: str
    name: str
    provider: str = Field(
        deprecated="This field is deprecated and will be removed in future versions.",
    )
    updated_at: ISO8601DatetimeUTCLike
    created_at: ISO8601DatetimeUTCLike

    def __str__(self):
        return f'DeviceSummary(name="{self.name}", id="{self.id}")'



from __future__ import annotations

import enum
import functools
import importlib
import importlib.metadata
import logging
import time
from collections import deque
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Self, cast

import numpy as np
import qblox_instruments as qbxi
from boulderopalscaleupsdk.device.controller import qblox as qbxs
from packaging.version import Version
from pydantic import TypeAdapter

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    import qcodes.instrument
    import qcodes.parameters
    from qblox_instruments.qcodes_drivers import module as qbxi_module
    from qblox_instruments.qcodes_drivers import sequencer as qbxi_sequencer

log = logging.getLogger(__name__)

QbxStackType = dict[str, qbxi.Cluster]

# ==================================================================================================
# compat
# ==================================================================================================
qblox_instruments_version = Version(importlib.metadata.version("qblox-instruments"))
if qblox_instruments_version >= Version("0.17.0"):

    def _get_sequencer_acquisition_status(seq: qbxi_sequencer.Sequencer) -> bool:
        # `check_seq_state` MUST be set to False otherwise, this will never return True.
        return seq.get_acquisition_status(timeout=0, check_seq_state=False)
else:

    def _get_sequencer_acquisition_status(seq: qbxi_sequencer.Sequencer) -> bool:
        return seq.get_acquisition_status(timeout=0)


# ==================================================================================================
# Instrument management
# ==================================================================================================


@functools.cache
def get_cluster(name: str, host: str, port: int | None = None) -> qbxi.Cluster:
    """
    Utility to get the Qblox cluster with try-catch.

    Parameters
    ----------
    name: str
        The name of the cluster
    host: str
        The address to connect to the cluster to
    port: int | None, optional
        The TCP port to communicate with the Cluster over. Defaults to None to use the default port.

    Returns
    -------
    qblox_instruments.Cluster
        The cluster

    Raises
    ------
    TimeoutError
    """
    try:
        log.debug("Resolving cluster %s from %s", name, host)
        cluster = qbxi.Cluster(name, host, port=port)
    except KeyError:
        log.warning("A cluster with the name '%s' already exists.", name)
        cluster = cast("qbxi.Cluster", qbxi.Cluster._all_instruments[name])
    except TimeoutError:  # pragma: no cover
        log.error("Timed out trying to connect to cluster %s @ %s", name, host)  # noqa: TRY400
        raise

    return cluster


@functools.cache
def get_cluster_modules(cluster: qbxi.Cluster) -> dict[qbxs.ModuleAddr, qbxi_module.Module]:
    """
    Get all connected modules for a QBLOX cluster.

    Parameters
    ----------
    cluster: qblox_instruments.Cluster
        The cluster

    Returns
    -------
    dict[qbxs.ModuleAddr, qblox_instruments.qcodes_drivers.module.Module]
        The connected modules using Q-CTRL addressing for modules as the key.
    """
    cluster_name = cluster.name
    connected_modules = cluster.get_connected_modules()
    return {qbxs.ModuleAddr(cluster_name, slot): mod for slot, mod in connected_modules.items()}


def get_module(stack: QbxStackType, mod_addr: qbxs.ModuleAddr) -> qbxi_module.Module:
    """
    Retrieve a single module from a QBLOX control stack.

    Parameters
    ----------
    stack: dict[str, qblox_instruments.Cluster]
        The control stack
    mod_addr: boulderopalscaleupsdk.device.controller.qblox.ModuleAddr
        The QBLOX module address

    Returns
    -------
    qblox_instruments.qcodes_drivers.module.Module
        The module if it is connected

    Raises
    ------
    ValueError
        - If the module is addressed to a QBLOX cluster not in the control stack
        - If the module is not connected to any QBLOX cluster in the control stack
    """
    cluster = stack.get(mod_addr.cluster)
    if cluster is None:
        msg = f"Stack does not have cluster {mod_addr.cluster}."
        log.error(msg)
        raise ValueError(msg)

    cluster_modules = get_cluster_modules(cluster)
    mod = cluster_modules.get(mod_addr)
    if mod is None:
        msg = f"Module {mod_addr!s} is not connected."
        log.error(msg)
        raise ValueError(msg)
    return mod


def get_module_types(stack: QbxStackType) -> dict[qbxs.ModuleAddr, qbxs.ModuleType]:
    """
    Get all the module types for each connected module in a stack.

    Parameters
    ----------
    stack: dict[str, qblox_instruments.Cluster]
        The control stack

    Returns
    -------
    dict
        The mapping of module addresses to module types
    """
    modules: dict[qbxs.ModuleAddr, qbxs.ModuleType] = {}
    for cluster in stack.values():
        modules |= {
            addr: _get_module_type(mod) for addr, mod in get_cluster_modules(cluster).items()
        }
    return modules


def _get_module_type(module: qbxi_module.Module) -> qbxs.ModuleType:
    if module.is_qcm_type:
        return qbxs.ModuleType.QCM_RF if module.is_rf_type else qbxs.ModuleType.QCM
    if module.is_qrm_type:
        return qbxs.ModuleType.QRM_RF if module.is_rf_type else qbxs.ModuleType.QRM
    is_qdm = getattr(module, "is_qdm_type", False)
    if is_qdm:
        return qbxs.ModuleType.QDM
    is_qtm = getattr(module, "is_qtm_type", False)
    if is_qtm:
        return qbxs.ModuleType.QTM
    is_eom = getattr(module, "is_eom_type", False)
    if is_eom:
        return qbxs.ModuleType.EOM
    is_linq = getattr(module, "is_linq_type", False)
    if is_linq:
        return qbxs.ModuleType.LINQ
    is_qrc = getattr(module, "is_qrc_type", False)
    if is_qrc:
        return qbxs.ModuleType.QRC

    raise NotImplementedError


# ==================================================================================================
# Arming
# ==================================================================================================


def connect_channel(ch: qbxs.ChannelType, module: qbxi_module.Module, seq_num: int) -> None:
    """
    Connect a sequencer to a channel.

    Parameters
    ----------
    ch: boulderopalscaleupsdk.device.controller.qblox.ChannelType
        The channel to connect to
    module: qblox_instruments.qcodes_drivers.module.Module
        The module the sequencer is located on
    seq_num: int
        The sequencer number
    """

    match ch:
        case qbxs.IQMixedChannel():
            _connect_real_channel(ch, module, seq_num)
        case qbxs.IQChannel():
            _connect_iq_channel(ch, module, seq_num)
        case qbxs.SingleChannel():
            _connect_single_channel(ch, module, seq_num)


def _connect_real_channel(
    ch: qbxs.IQMixedChannel,
    module: qbxi_module.Module,
    seq_num: int,
) -> None:
    seq = module.sequencers[seq_num]
    is_rf = module.is_rf_type
    if ch.port.direction == "out":
        f_name = f"connect_out{ch.port.number}"
        arg = "IQ" if is_rf else "I"
        log.debug("seq:%s -> ch:%s: with %s('%s')", seq.name, ch, f_name, arg)
        getattr(seq, f_name)(arg)
    else:
        conn = f"in{ch.port.number}"
        if is_rf:
            log.debug("seq:%s -> ch:%s: connect_acq('%s')", seq.name, ch, conn)
            seq.connect_acq(conn)
        else:
            log.debug("seq:%s -> ch:%s: connect_acq_I('%s')", seq.name, ch, conn)
            seq.connect_acq_I(conn)


def _connect_iq_channel(ch: qbxs.IQChannel, module: qbxi_module.Module, seq_num: int) -> None:
    seq = module.sequencers[seq_num]
    if ch.i_port.direction != ch.q_port.direction:
        raise ValueError("IQ ports of a single channel must be in the same direction.")
    direction = ch.i_port.direction
    if direction == "out":
        i_f = f"connect_out{ch.i_port.number}"
        q_f = f"connect_out{ch.q_port.number}"
        log.debug("seq:%s -> ch:%s: %s('I') + %s('Q')", seq.name, ch, i_f, q_f)
        getattr(seq, i_f)("I")
        getattr(seq, q_f)("Q")
    else:
        i_c = f"in{ch.i_port.number}"
        q_c = f"in{ch.q_port.number}"
        log.debug("seq:%s -> ch:%s: connect_acq_I(%s) + connect_acq_Q(%s)", seq.name, ch, i_c, q_c)
        seq.connect_acq_I(i_c)
        seq.connect_acq_Q(q_c)


def _connect_single_channel(
    ch: qbxs.SingleChannel,
    module: qbxi_module.Module,
    seq_num: int,
) -> None:
    seq = module.sequencers[seq_num]
    if module.is_rf_type:
        raise ValueError("Cannot use single channels for RF modules.")
    direction = ch.port.direction
    if direction == "out":
        f_name = f"connect_out{ch.port.number}"
        path = "I" if ch.path == 0 else "Q"
        getattr(seq, f_name)(path)
        log.debug("seq:%s -> ch:%s: %s(%s)", seq.name, ch, f_name, path)
    else:
        f_name = "connect_acq_I" if ch.path == 0 else "connect_acq_Q"
        port = f"in{ch.port.number}"
        log.debug("seq:%s -> ch:%s: %s(%s)", seq.name, ch, f_name, port)
        getattr(seq, f_name)(port)


def configure_instrument(
    instrument: qcodes.instrument.InstrumentBase,
    parameters: dict[str, Any],
) -> None:
    for name, val in parameters.items():
        delegate = instrument.parameters.get(name)
        if delegate is None:
            raise KeyError(f"Instrument {instrument.name} does not have parameter {name}.")
        delegate.set(val)


@dataclass
class ArmedSequencers:
    """
    Intermediate object tracking armed sequencers ready for execution.

    Attributes
    ----------
    sequencers: dict[SequenceAddr, Sequencer]
        The sequencers that are armed and ready for execution. Armed here means that the necessary
        QCoDeS sequencer parameters have been set, the correct connections have been made, the
        sequence program has been uploaded and the `.arm_sequencer()` API has been called.
    acquisitions_to_collect: set[SequenceAddr]
        The set of sequencers for which we need to collect acquisitions from.
    scopes_to_collect: dict[SequenceAddr, list[str]]
        The named scopes to collect for each SequenceAddr.
    """

    stack: dict[str, qbxi.Cluster]
    sequencers: dict[qbxs.SequencerAddr, qbxi_sequencer.Sequencer]
    acquisitions_to_collect: set[qbxs.SequencerAddr]
    scopes_to_collect: dict[qbxs.SequencerAddr, list[str]]


def calibrate_lo_leakage(
    prepared_program: qbxs.PreparedProgram,
    stack: QbxStackType,
    elements: list[str],
) -> None:
    """
    Run LO leakage calibration for the module/port pairs corresponding to the provided
    elements.

    See also:
    https://docs.qblox.com/en/main/tutorials/q1asm_tutorials/QRM-RF/110_automatic_mixer_calibration.html

    Parameters
    ----------
    prepared_program: boulderopalscaleupsdk.device.controller.qblox.PreparedProgram
        The program that we wish to execute
    stack: dict[str, qblox_instruments.Cluster]
        The QBLOX control stack to target
    elements: list[str]
        A list with the names of the elements to run the calibration on.
    """
    for elem in elements:
        seq_addr = prepared_program.sequence_programs[elem].sequencer_addr
        ch_out = prepared_program.sequence_programs[elem].ch_out
        mod_addr = seq_addr.module

        ports: list[qbxs.PortAddr] = []
        match ch_out:
            case qbxs.IQMixedChannel(port=port):
                ports.append(port)
            case qbxs.IQChannel(i_port=i_port, q_port=q_port):
                ports += [i_port, q_port]
            case _:
                raise RuntimeError(f"Unsupported channel {ch_out} for element {elem}.")

        log.info("Suppressing LO leakage for module: %s", mod_addr)
        module = get_module(stack, mod_addr)
        module_type = _get_module_type(module)
        match module_type:
            case qbxs.ModuleType.QCM_RF:
                for port in ports:
                    if port.number == 0:
                        module.out0_lo_cal()
                    elif port.number == 1:
                        module.out1_lo_cal()
                    else:
                        raise RuntimeError(
                            f"Invalid number for port {port}. QCM-RF only has two ports.",
                        )
            case qbxs.ModuleType.QRM_RF:
                module.out0_in0_lo_cal()
            case _:
                log.warning("Cannot calibrate LO for module %s, it does not have LO.", mod_addr)


def arm_sequencers(  # noqa: C901, PLR0912, PLR0915
    prepared_program: qbxs.PreparedProgram,
    stack: QbxStackType,
    calibrate_elements: list[str] | None = None,
    reset: bool = False,
) -> ArmedSequencers:
    """
    Arm sequences from a prepared program.

    Parameters
    ----------
    prepared_program: boulderopalscaleupsdk.device.controller.qblox.PreparedProgram
        The program that we wish to execute
    stack: dict[str, qblox_instruments.Cluster]
        The QBLOX control stack to target
    calibrate_elements: list[str] or None, optional
        A list with the names of the elements to run mixer calibration on.
    reset: bool, optional
        When set, will reset each cluster in the stack. Defaults to True.

    Returns
    -------
    ArmedSequencers
        A data class describing the armed sequencers ready for execution

    Raises
    ------
    RuntimeError
        If a targeted sequencer is in an invalid state.
    ValueError
        If a sequence program is invalid (e.g. due to a syntax error)
    """
    if reset:
        for cluster_name, cluster in stack.items():
            log.debug("Resetting cluster:%s", cluster_name)
            cluster.reset()

    modules = {addr: get_module(stack, addr) for addr in prepared_program.modules}

    sequencers_to_calibrate: list[qbxs.SequencerAddr] = []

    if calibrate_elements:
        sequencers_to_calibrate = [
            prepared_program.sequence_programs[elem].sequencer_addr for elem in calibrate_elements
        ]

    for mod_addr, mod in modules.items():
        log.debug("Disconnecting mod:%s connections and stopping sequencers", mod_addr)
        mod.disconnect_outputs()
        if mod.is_qrm_type:
            mod.disconnect_inputs()
        mod.stop_sequencer()
        mod_prep = prepared_program.modules[mod_addr]
        mod_params = mod_prep.params.model_dump(exclude_unset=True, exclude_none=True)
        for param_key, param_val in mod_params.items():
            log.debug("Configured module %s: %s=%s", mod_addr, param_key, param_val)
        configure_instrument(mod, mod_params)

    if calibrate_elements:
        calibrate_lo_leakage(
            prepared_program=prepared_program,
            stack=stack,
            elements=calibrate_elements,
        )

    sequencers: dict[qbxs.SequencerAddr, qbxi_sequencer.Sequencer] = {}
    acquisitions_to_collect: set[qbxs.SequencerAddr] = set()
    scopes_to_collect: dict[qbxs.SequencerAddr, list[str]] = {}
    for ref, prepared in prepared_program.sequence_programs.items():
        seq_addr = prepared.sequencer_addr
        ch_out = prepared.ch_out
        ch_in = prepared.ch_in
        mod = modules[seq_addr.module]
        seq_prog = prepared.sequence_program

        seq: qbxi_sequencer.Sequencer = cast(
            "qbxi_sequencer.Sequencer",
            mod.sequencers[seq_addr.number],
        )

        # Check status
        status = seq.get_sequencer_status(timeout=0)
        if status.state not in [qbxi.SequencerStates.IDLE, qbxi.SequencerStates.STOPPED]:
            message = f"Sequencer {seq.name} in invalid state: {status.state}."
            log.exception(message)
            # TODO: Consider changing exception class here; a RuntimeError may not be very helpful
            raise RuntimeError(message)

        # Clear status flags so we have a clean state management
        seq.clear_sequencer_flags()

        # Configure
        log.info(
            "Preparing seq:%s for prog:%s targetting ch_out:%s, ch_in:%s",
            seq_addr,
            ref,
            ch_out,
            ch_in,
        )
        try:
            data = seq_prog.sequence_data()
            if data is not None:
                seq.sequence(data)
        except RuntimeError as exc:
            # TODO: Improve error reporting to help debug program errors.
            log.exception("Failed to upload prog:%s to seq:%s.", ref, seq_addr)
            raise ValueError(f"Invalid program: {exc!s}.") from exc

        connect_channel(ch_out, mod, seq_addr.number)
        if ch_in:
            connect_channel(ch_in, mod, seq_addr.number)

        seq_params = prepared.sequence_program.params.model_dump(
            exclude_unset=True,
            exclude_none=True,
        )
        for param_key, param_val in seq_params.items():
            log.debug("Configured sequencer %s: %s=%s", seq_addr, param_key, param_val)
        configure_instrument(seq, seq_params)

        if len(seq_prog.acquisitions) > 0:
            if not mod.is_qrm_type:
                log.warning(
                    "Sequence %s on %s has acquisitions but module cannot readout",
                    ref,
                    seq_addr,
                )
            else:
                acquisitions_to_collect.add(seq_addr)
                if seq_prog.acquisition_scopes:
                    scopes_to_collect[seq_addr] = seq_prog.acquisition_scopes

        if seq_addr in sequencers_to_calibrate:
            log.info("Suppressing undesired sideband for sequencer: %s", seq_addr)
            seq.sideband_cal()

        # Arm

        if seq_prog.params_only:
            continue
        sequencers[seq_addr] = seq

    # NOTE: We must arm sequencers AFTER all configurations, since configurations will disarm
    #       previously armed sequencers on the same instrument.
    log.info("Arming sequencers...")
    for seq in sequencers.values():
        seq.arm_sequencer()
    log.info("%d sequencers armed", len(sequencers))

    return ArmedSequencers(
        stack=stack,
        sequencers=sequencers,
        acquisitions_to_collect=acquisitions_to_collect,
        scopes_to_collect=scopes_to_collect,
    )


# ==================================================================================================
# Execution
# ==================================================================================================
class ExecState(enum.Enum):
    DONE = enum.auto()
    ERRORED = enum.auto()
    IN_PROGRESS = enum.auto()
    INVALID_STATE = enum.auto()


def _get_acquisitions_status(seq_addr: qbxs.SequencerAddr, seq: qbxi_sequencer.Sequencer) -> bool:
    try:
        return _get_sequencer_acquisition_status(seq)  # see compat
    except TimeoutError:
        log.debug("seq:%s timed out on get_acquisition_status", seq_addr)
        return False
    except NotImplementedError:  # Not a QRM module
        log.warning("seq:%s does not have any acquisitions but was requested to check", seq_addr)
        return False  # TODO: Consider raising an error / returning invalid state here


def _get_exec_state(
    seq_addr: qbxs.SequencerAddr,
    seq: qbxi_sequencer.Sequencer,
    has_acquisitions: bool = False,
) -> ExecState:
    """
    Map a sequencer's state to ExecState.

    Parameters
    ----------
    seq_addr: boulderopalscaleupsdk.device.controller.qblox.SequencerAddr
        The sequencer's address; used for logging
    seq: qblox_instruments.qcodes_drivers.sequencer.Sequencer
        The sequencer QCoDeS driver instance
    has_acquisitions: bool, optional
        When set, this function will check the sequencer's acquisition status.

    Returns
    -------
    ExecState
        The sequencer execution state
    """
    seq_status: qbxi.SequencerStatus = seq.get_sequencer_status(timeout=0)
    log.debug("seq:%s sequencer status %r", seq_addr, seq_status)

    # -------------------------------------------------------------------------
    # The following checks will return if state can be determined, otherwise pass.
    # -------------------------------------------------------------------------
    # Check status
    status = cast("qbxi.SequencerStatuses", seq_status.status)
    match status:
        case qbxi.SequencerStatuses.ERROR:
            log.error("seq:%s exec done with errors: %s", seq_addr, seq_status.err_flags)
            return ExecState.ERRORED
        case qbxi.SequencerStatuses.WARNING:
            log.warning("seq:%s has warnings: %s", seq_addr, seq_status.warn_flags)
            # pass
        case qbxi.SequencerStatuses.OKAY:
            pass

    # ---------------------------------
    # Check state
    state = cast("qbxi.SequencerStates", seq_status.state)
    match state:
        case qbxi.SequencerStates.IDLE | qbxi.SequencerStates.ARMED:
            log.error("seq:%s not armed", seq_addr)
            return ExecState.INVALID_STATE
        case qbxi.SequencerStates.RUNNING:
            return ExecState.IN_PROGRESS
        case qbxi.SequencerStates.Q1_STOPPED | qbxi.SequencerStates.STOPPED:
            pass

    # ---------------------------------
    # Acquisitions
    if not has_acquisitions:
        return ExecState.DONE

    if qbxi.SequencerStatusFlags.DISARMED in seq_status.info_flags:
        # TODO: Need to clarify behaviour. See: https://qctrl.atlassian.net/browse/SCUP-1203
        #       It appears if the sequencer is disarmed then checking the acquisition status will
        #       hang indefinitely; returning DONE will still give valid results.
        log.warning("seq:%s was disarmed", seq_addr)
        return ExecState.DONE

    acquisitions_ready = _get_acquisitions_status(seq_addr, seq)
    return ExecState.DONE if acquisitions_ready else ExecState.IN_PROGRESS


def _poll_and_iter_ready_sequencers(  # noqa: C901
    armed_sequencers: ArmedSequencers,
    timeout_poll_res: float = 0.5,
    timeout: float = 30,
) -> Iterator[tuple[qbxs.SequencerAddr, qbxi_sequencer.Sequencer]]:
    """
    Poll a set of armed sequencers yielding each time a single sequencer has finished.

    Parameters
    ----------
    armed_sequencers: ArmedSequencers
        The sequencers to poll; these sequencers must be executed before calling this function,
        otherwise a RuntimeError will be raised
    timeout_poll_res: float, optional
        The time (in seconds) to wait before each iteration of polling. Defaults to 0.5 seconds
    timeout: float, optional
        The total time (in seconds) to poll for until raising a TimeoutError. Defaults to 30 seconds

    Yields
    ------
    tuple[
        boulderopalscaleupsdk.device.controller.qblox.SequencerAddr,
        qblox_instruments.qcodes_drivers.sequencer.Sequencer
    ]
        Yields a ready sequencers as they are found to be ready.

    Raises
    ------
    RuntimeError
        If any sequencer in the armed set is in an invalid state or errored during a run
    TimeoutError
        If a timeout is set and the total elapsed time has exceeded that value

    Notes
    -----
    Each poll on each sequencer is a blocking call as there is limited async support for QBLOX.
    """
    t_start = time.time()
    not_ready = list(armed_sequencers.sequencers.keys())
    error_count = 0

    while True:
        not_ready2 = []
        for seq_addr in not_ready:
            seq = armed_sequencers.sequencers[seq_addr]
            has_acquisitions = seq_addr in armed_sequencers.acquisitions_to_collect
            try:
                exec_state = _get_exec_state(seq_addr, seq, has_acquisitions)
            except TimeoutError:
                # NOTE: qblox-instruments SCPI uses IEEE488.2 over an AF_INET socket. As of 0.12.0
                #       this socket is configured with a 60 second timeout after connection and we
                #       have observed status checks to timeout over a proxied connection. This
                #       disrupts our timeout configuration and leads to a poor user-experience due
                #       to conflicting timeout configurations. Instead, we will simply warn the user
                #       about failed connections and move on, relying on our simple timer.
                error_count += 1
                log.warning(
                    "Sequencer:%s not responding to status check, ignoring. Error count %d...",
                    seq_addr,
                    error_count,
                )
                exec_state = ExecState.IN_PROGRESS
            else:
                match exec_state:
                    case ExecState.DONE:
                        yield seq_addr, seq
                    case ExecState.IN_PROGRESS:
                        not_ready2.append(seq_addr)
                    case ExecState.ERRORED | ExecState.INVALID_STATE:
                        for seq_addr_ in not_ready:
                            _cleanup_sequencer_config(
                                armed_sequencers.sequencers[seq_addr_],
                                seq_addr_,
                            )
                        raise RuntimeError(f"Sequencer:{seq_addr} has execution errors.")

        if not not_ready2:
            break
        not_ready = not_ready2

        elapsed = time.time() - t_start
        log.debug("polling sequencers, %.2f seconds elapsed", elapsed)
        if timeout and elapsed >= timeout:
            for seq_addr in not_ready:
                seq = armed_sequencers.sequencers[seq_addr]
                seq.stop_sequencer()
                _cleanup_sequencer_config(seq, seq_addr)
            raise TimeoutError

        time.sleep(timeout_poll_res)


def _cleanup_sequencer_config(
    sequencer: qbxi_sequencer.Sequencer,
    sequencer_addr: qbxs.SequencerAddr,
):
    """
    Clean up sequencer configuration after the end of the execution, so they don't
    interfere with subsequent programs.
    """
    if sequencer.sync_en():
        sequencer.sync_en(False)
        log.info("removed syncing flag for seq:%s", sequencer_addr)


def execute_armed_sequencers(
    armed: ArmedSequencers,
    timeout_poll_res: float = 1,
    timeout: float = 30,
) -> dict[qbxs.SequencerAddr, qbxs.OutputSequencerAcquisitions]:
    """
    Execute against ArmedSequencers.

    Parameters
    ----------
    armed: ArmedSequencers
        The ArmedSequencers to target. see `arm_sequencers`.
    timeout_poll_res: float, optional
        The polling resolution (in seconds); the amount of time to wait between each poll iteration.
    timeout: float, optional
        The time (in seconds) before a TimeoutError should be raised. If set to zero, this will poll
        indefinitely

    Returns
    -------
    dict[qbxs.SequencerAddr, qbxs.OutputSequencerAcquisitions]
        The execution results for each sequencer.

    Raises
    ------
    TimeoutError
        If `timeout` is set, and the polling time exceeds the configured value.
    RuntimeError
        If a sequencer encounters a run-error or enters an invalid state.
    """
    # Run sequencers
    for seq_addr, seq in armed.sequencers.items():
        try:
            seq.start_sequencer()
        except RuntimeError:
            log.exception("Failed to start sequencer %s", seq_addr)
            if cluster := armed.stack.get(seq_addr.cluster):
                log.info(
                    "Module %s failed with assembler log: %s",
                    seq_addr.module,
                    cluster.get_assembler_log(seq_addr.module.slot),  # pyright: ignore[reportAttributeAccessIssue]
                )
            raise

        log.info("started seq:%s", seq_addr)

    output_type_adapter = TypeAdapter(qbxs.OutputSequencerAcquisitions)

    results: dict[qbxs.SequencerAddr, qbxs.OutputSequencerAcquisitions] = {}
    for ready_seq_addr, ready_seq in _poll_and_iter_ready_sequencers(
        armed,
        timeout_poll_res=timeout_poll_res,
        timeout=timeout,
    ):
        try:
            if ready_seq_addr in armed.acquisitions_to_collect:
                for scope in armed.scopes_to_collect.get(ready_seq_addr, []):
                    ready_seq.store_scope_acquisition(scope)
                acquisitions = ready_seq.get_acquisitions()
                ready_seq.delete_acquisition_data(all=True)
                results[ready_seq_addr] = output_type_adapter.validate_python(acquisitions)
        finally:
            ready_seq.stop_sequencer()
            log.info("seq:%s done and stopped", ready_seq_addr)
            _cleanup_sequencer_config(ready_seq, ready_seq_addr)

    return results


def expand_and_label_results(
    prepared_program: qbxs.PreparedProgram,
    results: dict[qbxs.SequencerAddr, qbxs.OutputSequencerAcquisitions],
) -> dict[str, qbxs.OutputAcquisition]:
    """
    Utility function to re-index and expand the results from execution.

    Parameters
    ----------
    prepared_program: PreparedProgram
        The program that defines the sequence_programs and acquisitions of each sequence_program.
    results: dict[qbxs.SequencerAddr, qbxs.OutputSequencerAcquisitions]
        The results from executing against an armed sequencer.

    Returns
    -------
    dict[str, OutputAcquisition]
        The output acquisitions indexed with string keys following the format `<program>_<acq>`,
        where `<program>` corresponds to the program key in `PreparedProgram.sequence_programs` and
        `<acq>` corresponds to the acquisition key in `SequenceProgram.acquisitions`.
    """
    seq_prog_map: dict[qbxs.SequencerAddr, str] = {
        seq_prog.sequencer_addr: name
        for name, seq_prog in prepared_program.sequence_programs.items()
    }
    return {
        f"{seq_prog_map[addr]}_{acq_ref}": acq.acquisition
        for addr, result in results.items()
        for acq_ref, acq in result.items()
    }


# ==================================================================================================
# Misc
# ==================================================================================================
@dataclass
class _SetpointSweep:
    """
    Utility method to sweep a parameter delegate over a list of values over t_delta time steps.
    """

    instrument: qcodes.instrument.InstrumentBase
    delegate: qcodes.parameters.ParameterBase
    values: Iterable[float]
    t_delta: float = 0.5

    _last_emit: float | None = field(default=None, init=False)
    _values_queue: deque[float] = field(init=False)

    def __post_init__(self) -> None:
        self._values_queue = deque(self.values)

    @classmethod
    def get_ramp(
        cls,
        instrument: qcodes.instrument.InstrumentBase,
        param: str,
        target: float,
        step: float,
        t_delta: float = 0.5,
    ) -> Self | None:
        delegate = instrument.parameters.get(param)
        if delegate is not None and abs(curr := delegate.get()) > 0:
            step_sign = 1 if target > curr else -1
            return cls(
                instrument,
                delegate,
                np.arange(curr, target, step_sign * step),
                t_delta,
            )
        return None

    def step(self, t_now: float) -> float:
        """
        Step to the next set point if t_now meets the deadline.

        Parameters
        ----------
        t_now: float
            The current time. A consistent unit must be used across all method calls for this to
            work correctly

        Returns
        -------
        float: the remaining time until the next set point should be set.
        """
        if not self._values_queue:
            raise StopIteration
        if self._last_emit is None:
            self._last_emit = t_now
        elif (remaining := self._last_emit + self.t_delta - t_now) > 0:
            return remaining
        val = self._values_queue.popleft()
        log.debug("Setting %s param '%s' = %.2f", self.instrument.name, self.delegate.name, val)
        self.delegate.set(val)

        self._last_emit = t_now
        return self.t_delta


def _run_setpoint_sweeps(sweeps: list[_SetpointSweep]) -> None:
    """
    Run the sweeps using a work queue strategy.

    The work queue allows setting one setpoint per instrument. Once a sweep is stepped, the wait
    time is indexed against the target instrument; other sweeps targeting the same instrument will
    be blocked until this wait time has been elapsed.
    """
    waiters: dict[str, float] = {}
    queue: deque[_SetpointSweep] = deque(sweeps)

    while True:
        n_items = len(queue)
        if n_items == 0:
            break
        for _ in range(n_items):
            sw = queue.popleft()
            key = sw.instrument.name
            if waiters.get(key, 0) > 0:
                queue.append(sw)
            else:
                try:
                    waiters[key] = sw.step(time.time())
                except StopIteration:
                    pass
                else:
                    queue.append(sw)

        if waiters:  # Clear all wait times by taking the largest sleep time.
            time.sleep(max(waiters.values()))
        waiters = {}


def reset_with_ramp_down(cluster: qbxi.Cluster, t_delta: float = 0.5) -> None:
    """
    Reset cluster with a ramp down.

    Parameters
    ----------
    t_delta: float, optional
        The time step between set points for a single module.
    """
    modules = get_cluster_modules(cluster)

    # Collect all the required set point sweeps
    sweeps: list[_SetpointSweep] = []
    for mod_addr, mod in modules.items():
        params: list[tuple[str, float]]
        match _get_module_type(mod):
            case qbxs.ModuleType.QCM_RF | qbxs.ModuleType.QRM_RF:
                params = [("offset_awg_path0", 0.2), ("offset_awg_path1", 0.2)]
            case qbxs.ModuleType.QCM:
                params = [
                    ("out0_offset", 0.2),
                    ("out1_offset", 0.2),
                    ("out2_offset", 0.2),
                    ("out3_offset", 0.2),
                ]
            case qbxs.ModuleType.QRM:
                params = [("out0_offset", 0.2), ("out1_offset", 0.2)]
            case _:
                continue

        for pp, step in params:
            if ss := _SetpointSweep.get_ramp(mod, pp, 0, step, t_delta=t_delta):
                log.debug("Ramping '%s' for %s: %s", pp, mod_addr, ss.values)
                sweeps.append(ss)

    # Run the sweeps
    _run_setpoint_sweeps(sweeps)

    # Reset the cluster
    log.info("Resetting cluster %s", cluster.name)
    cluster.reset()



"""
Version module for boulder-opal-scale-up package.

This module uses the centralized version utilities from boulder-opal-scale-up-sdk
to provide version information in a format compatible with the poetry-get-version workflow.
"""

import argparse
import sys

from boulderopalscaleupsdk.utils.version import get_version


def main():
    """Main entry point for version script."""
    parser = argparse.ArgumentParser(description="Get package version")
    parser.add_argument(
        "--strip-local",
        action="store_true",
        help="Strip local version identifiers",
    )

    args = parser.parse_args()

    try:
        version = get_version(
            package_name="boulder-opal-scale-up",
            strip_local=args.strip_local,
        )
        print(version)  # noqa: T201
        return 0  # noqa: TRY300
    except ValueError as e:
        print(f"Error: {e}", file=sys.stderr)  # noqa: T201
        return 1


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

__all__ = [
    "Controller",
    "OPX1000Controller",
    "OPXController",
    "QBLOXController",
]

from boulderopalscaleup.controllers._base import Controller
from boulderopalscaleup.controllers.qblox import QBLOXController
from boulderopalscaleup.controllers.quantum_machines.opx import OPXController
from boulderopalscaleup.controllers.quantum_machines.opx1000 import OPX1000Controller



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

import logging
from typing import Self

import numpy as np
import qblox_instruments as qbxi
from boulderopalscaleupsdk.device.controller import qblox as qbxs
from boulderopalscaleupsdk.protobuf.v1 import agent_pb2
from google.protobuf.struct_pb2 import Struct

from boulderopalscaleup import qblox as qbxc
from boulderopalscaleup.controllers import Controller

LOG = logging.getLogger(__name__)


class QBLOXController(Controller):
    """
    QBLOX Controller.

    Parameters
    ----------
    stack: dict[str, Cluster]
        The control stack with all connected clusters. The key of the stack dictionary should match
        the cluster identifier.
    timeout: float, optional
        The timeout in seconds before a program should be interrupted and exited. Defaults to 30s.
    reset: bool, optional
        When set, will reset all clusters on the stack on initialization. Defaults to False.
    """

    def __init__(self, stack: dict[str, qbxi.Cluster], timeout: float = 30, reset: bool = False):
        self._stack = stack
        if reset:
            for cluster in stack.values():
                cluster.reset()
        self._timeout = timeout

    @classmethod
    def new(cls, *cluster_hosts: tuple[str, str]) -> Self:
        stack = {name: qbxc.get_cluster(name, host) for name, host in cluster_hosts}
        return cls(stack)

    async def run_program(
        self,
        program_request: agent_pb2.RunProgramRequest,
    ) -> agent_pb2.RunProgramResponse:
        program = qbxs.PreparedProgram.loads(program_request.program)
        if LOG.isEnabledFor(logging.DEBUG):
            for ch, psp in program.sequence_programs.items():
                LOG.debug(
                    "Running program for '%s' on ch_out=%s\n %s",
                    psp.ch_out,
                    ch,
                    psp.sequence_program.program,
                )
        armed = qbxc.arm_sequencers(
            prepared_program=program,
            stack=self._stack,
            calibrate_elements=list(program_request.calibrate_elements),
            reset=False,
        )
        exec_results = qbxc.execute_armed_sequencers(armed, timeout=self._timeout)
        raw_flattened = self._flatten_sequencer_results(program, exec_results)
        post_processed = self._results_post_process(raw_flattened)
        raw_data_struct = Struct()
        raw_data_struct.update(post_processed)
        return agent_pb2.RunProgramResponse(raw_data=raw_data_struct)

    @staticmethod
    def _results_post_process(raw: qbxs.SequencerResults) -> dict[str, list[float]]:
        ret = {}
        for result_key, result_bins in raw.bins.items():
            ret[f"{result_key}_i"] = np.real(result_bins).tolist()
            ret[f"{result_key}_q"] = np.imag(result_bins).tolist()
        return ret

    @staticmethod
    def _flatten_sequencer_results(
        prepared: qbxs.PreparedProgram,
        execution_output: dict[qbxs.SequencerAddr, qbxs.OutputSequencerAcquisitions],
    ) -> qbxs.SequencerResults:
        bins: dict[str, np.ndarray] = {}
        scopes: dict[str, np.ndarray] = {}
        for seq_addr, seq_output in execution_output.items():
            seq_prog = prepared.get_sequencer_program(seq_addr)
            seq_results = qbxs.process_sequencer_output(seq_prog, seq_output)
            bins |= seq_results.bins
            scopes |= seq_results.scopes
        return qbxs.SequencerResults(scopes=scopes, bins=bins)

    async def run_mixer_calibration(
        self,
        calibration_request: agent_pb2.RunQuantumMachinesMixerCalibrationRequest,  # noqa: ARG002
    ) -> agent_pb2.RunQuantumMachinesMixerCalibrationResponse:
        LOG.warning(
            "Mixer calibration is not implemented for QBLOX controller. Skipping calibration.",
        )
        return agent_pb2.RunQuantumMachinesMixerCalibrationResponse(success=True)



from boulderopalscaleupsdk.common.dtypes import Duration
from boulderopalscaleupsdk.device.common import Component
from boulderopalscaleupsdk.device.processor.common import ComponentParameter
from pydantic import TypeAdapter
from pydantic.dataclasses import dataclass
from rich.columns import Columns
from rich.console import Console
from rich.table import Table


def _split_table_sizes(length: int) -> list[int]:
    # We split the parameters into sets of up to preferred_width columns,
    # each one will be displayed in a separate table.
    # (If there'd be a spare column, we add it to the first table.)
    preferred_width = 4
    if length <= preferred_width + 1:
        return [length]
    set_sizes = [preferred_width] * (length // preferred_width)
    remainder = length - sum(set_sizes)
    if remainder == 1:
        set_sizes[0] += 1
    elif remainder != 0:
        set_sizes.append(remainder)
    return set_sizes


@dataclass
class _DisplayInfo:
    label: str
    unit: str
    scale: float

    def _get_value(self, value: float | Duration | None) -> float | None:
        if isinstance(value, float | int):
            return value * self.scale
        if isinstance(value, Duration):
            return value.to_seconds() * self.scale
        return None

    @property
    def table_header(self) -> str:
        return f"{self.label}\n({self.unit})"

    def get_table_info(self, parameter: ComponentParameter | None) -> tuple[str, str, str]:
        if parameter is None:
            return "None", "None", "None"
        value = self._get_value(parameter.value)
        em = self._get_value(parameter.err_minus)
        ep = self._get_value(parameter.err_plus)
        std = em + ep if em is not None and ep is not None else None
        return str(value), str(std), str(parameter.calibration_status)


def display_element(reference: str, component: Component) -> None:
    console = Console()

    display_info = {
        k: TypeAdapter(_DisplayInfo).validate_python(extra["display"])
        for k, v in type(component).model_fields.items()
        if isinstance(extra := v.json_schema_extra, dict) and "display" in extra
    }
    headers = list(display_info.keys())

    tables = []
    for length in _split_table_sizes(len(display_info)):
        table = Table()
        table.add_column("", justify="right")
        values = ["value"]
        stds = ["std"]
        calibration_statuses = ["calibration status"]
        for _ in range(length):
            header = headers.pop(0)
            display = display_info[header]
            parameter = getattr(component, header)

            table.add_column(display.table_header, justify="center")
            value, std, calibration_status = display.get_table_info(parameter)
            values.append(value)
            stds.append(std)
            calibration_statuses.append(calibration_status)

        table.add_row(*values)
        table.add_row(*stds)
        table.add_row(*calibration_statuses)
        tables.append(table)

    console.print(Columns(tables, title=f"{type(component).__name__} {reference}"))



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from qctrlclient import ApiKeyAuth
from qctrlclient.defaults import (
    get_default_api_url,
)
from qctrlclient.exceptions import GraphQLQueryError


def get_default_api_key_auth(api_key: str) -> ApiKeyAuth:
    """
    Return a token-based authentication handler
    pointed to the default API URL.
    """
    auth = ApiKeyAuth(get_default_api_url(), api_key)

    # Check the API key.
    # We can thus infer that the API key is invalid if the access token cannot be fetched.
    try:
        _ = auth.access_token
    except GraphQLQueryError as error:
        raise RuntimeError(
            f"Invalid API key ({api_key}). Please check your key "
            "or visit https://accounts.q-ctrl.com to generate a new one.",
        ) from error

    return auth



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

__all__ = [
    "T1",
    "T2",
    "CZSpectroscopyByBias",
    "Chi01Scan",
    "DragLeakageCalibration",
    "FineAmplitudeCalibration",
    "OneQubitRB",
    "PowerRabi",
    "Ramsey",
    "ReadoutClassifier",
    "ResonatorSpectroscopy",
    "ResonatorSpectroscopyByBias",
    "ResonatorSpectroscopyByPower",
    "T2Echo",
    "TransmonAnharmonicity",
    "TransmonSpectroscopy",
    "ZZRamsey",
]


from boulderopalscaleupsdk.experiments import (
    T1,
    T2,
    Chi01Scan,
    CZSpectroscopyByBias,
    DragLeakageCalibration,
    FineAmplitudeCalibration,
    OneQubitRB,
    PowerRabi,
    Ramsey,
    ReadoutClassifier,
    ResonatorSpectroscopy,
    ResonatorSpectroscopyByBias,
    ResonatorSpectroscopyByPower,
    T2Echo,
    TransmonAnharmonicity,
    TransmonSpectroscopy,
    ZZRamsey,
)



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

__all__ = [
    "CWSIterable",
    "ConstantWaveform",
    "Duration",
    "Frequency",
    "GaussianWaveform",
    "HypIterable",
    "JobHistorySortOrder",
    "LinspaceIterable",
    "LogspaceIterable",
    "RangeIterable",
    "TimeUnit",
    "Waveform",
]


from boulderopalscaleupsdk.common.dtypes import Duration, Frequency, JobHistorySortOrder, TimeUnit
from boulderopalscaleupsdk.experiments import (
    ConstantWaveform,
    CWSIterable,
    GaussianWaveform,
    HypIterable,
    LinspaceIterable,
    LogspaceIterable,
    RangeIterable,
    Waveform,
)



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

import contextlib
import json
import logging
from typing import TYPE_CHECKING, cast

import numpy as np
import qm
from boulderopalscaleupsdk.device.controller.quantum_machines import QuaProgram
from boulderopalscaleupsdk.protobuf.v1 import agent_pb2
from google.protobuf.struct_pb2 import Struct
from pydantic.main import IncEx
from qm.jobs.running_qm_job import RunningQmJob

from boulderopalscaleup.controllers import Controller

if TYPE_CHECKING:
    from iqcc_cloud_client.qmm_cloud import (
        CloudJob,
        CloudQuantumMachine,
        CloudQuantumMachinesManager,
    )


# Import IQCC Cloud classes if available
with contextlib.suppress(ImportError):
    from iqcc_cloud_client.qmm_cloud import (
        CloudJob,
        CloudQuantumMachine,
        CloudQuantumMachinesManager,
    )

LOG = logging.getLogger(__name__)


class OPX1000Controller(Controller):
    def __init__(self, qmm: "qm.QuantumMachinesManager | CloudQuantumMachinesManager"):
        self.qmm = qmm

    async def run_program(
        self,
        program_request: agent_pb2.RunProgramRequest,
    ) -> agent_pb2.RunProgramResponse:
        LOG.info("Running experiment task %s", program_request)

        qua_program = QuaProgram.loads(program_request.program)

        # Defined as IncEx so linter is happy.
        # Exclude fields that are not supported by QOP 3.5
        exclude_fields: IncEx = cast(
            IncEx,
            {
                "qm_version": True,
                "controllers": {
                    "__all__": {
                        "fems": {
                            "__all__": {
                                "analog_outputs": {
                                    "__all__": {
                                        "filter": {
                                            "feedback": True,
                                        },
                                    },
                                },
                            },
                        },
                    },
                },
            },
        )
        config = qua_program.config.root.model_dump_json(exclude=exclude_fields, exclude_none=True)

        qm_instance = self._initialize_qm(config)

        LOG.info("Executing program.")
        match qm_instance:
            case CloudQuantumMachine():
                qm_job = qm_instance.execute(
                    qm.Program(program=qua_program.program),
                    options={"timeout": 600},
                )
            case _:
                qm_job = qm_instance.execute(
                    qm.Program(program=qua_program.program),
                )

        LOG.info("Handling results.")
        measurement_data = self._fetch_job_results(qm_job)

        def _convert(array):
            return np.asarray(array).astype(float).tolist()

        raw_data = {k: _convert(v) for k, v in measurement_data.items()}

        raw_data_struct = Struct()
        raw_data_struct.update(raw_data)
        return agent_pb2.RunProgramResponse(raw_data=raw_data_struct)

    async def run_mixer_calibration(
        self,
        calibration_request: agent_pb2.RunQuantumMachinesMixerCalibrationRequest,  # noqa: ARG002
    ) -> agent_pb2.RunQuantumMachinesMixerCalibrationResponse:
        """
        Run a mixer calibration on the device.
        """

        # No-op for OPX1000 because it does not need mixer calibration
        return agent_pb2.RunQuantumMachinesMixerCalibrationResponse(
            success=True,
        )

    def _initialize_qm(
        self,
        config_json: str,
    ) -> qm.QuantumMachine | CloudQuantumMachine:
        """
        Initialize a Quantum Machine from a config JSON string and export the config.
        """
        LOG.info("Initializing QM.")
        qua_config = json.loads(config_json)
        return self.qmm.open_qm(config=qua_config, close_other_machines=True)  # type: ignore[return-value]

    def _fetch_job_results(self, job: RunningQmJob | CloudJob):
        def _format(data):
            if (
                type(data) is np.ndarray
                and data.ndim > 0
                and type(data[0]) is np.void
                and len(data.dtype.names or []) == 1
            ):
                data = data["value"]
            return data

        result_handles = job.result_handles
        result_handles.wait_for_all_values()

        results = {}
        for data in list(result_handles.keys()):
            if hasattr(result_handles, data) and data != "__fridge_info":
                results[data] = _format(result_handles.get(data).fetch_all())  # type: ignore[union-attr]

        return results



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
import abc

from boulderopalscaleupsdk.protobuf.v1 import agent_pb2


class Controller(abc.ABC):
    @abc.abstractmethod
    async def run_program(
        self,
        program_request: agent_pb2.RunProgramRequest,
    ) -> agent_pb2.RunProgramResponse:
        raise NotImplementedError

    @abc.abstractmethod
    async def run_mixer_calibration(
        self,
        calibration_request: agent_pb2.RunQuantumMachinesMixerCalibrationRequest,
    ) -> agent_pb2.RunQuantumMachinesMixerCalibrationResponse:
        raise NotImplementedError



# Copyright 2025 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Client for the Boulder Opal Scale Up API.
"""

import json
import logging
import os
from dataclasses import dataclass
from importlib import metadata
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

import grpc
from boulderopalscaleupsdk.agent import Agent, AgentSettings, TaskHandler
from boulderopalscaleupsdk.common.dtypes import (
    DEFAULT_JOB_HISTORY_PAGE,
    DEFAULT_JOB_HISTORY_PAGE_SIZE,
    DEFAULT_JOB_HISTORY_SORT_ORDER,
    JobData,
    JobHistorySortOrder,
    JobId,
    JobSummary,
    SyncClientInterceptor,
)
from boulderopalscaleupsdk.constants import (
    DEFAULT_GRPC_MAX_RECEIVE_MESSAGE_LENGTH,
    DEFAULT_GRPC_MAX_SEND_MESSAGE_LENGTH,
)
from boulderopalscaleupsdk.device.config_loader import DeviceConfigLoader
from boulderopalscaleupsdk.device.controller.resolver import ControllerResolverService
from boulderopalscaleupsdk.device.defcal import DefCalData
from boulderopalscaleupsdk.device.processor import SuperconductingProcessor
from boulderopalscaleupsdk.device.processor.superconducting_processor import Resonator, Transmon
from boulderopalscaleupsdk.errors import ScaleUpServerError
from boulderopalscaleupsdk.experiments import Experiment
from boulderopalscaleupsdk.experiments.classifiers import ClassifierData
from boulderopalscaleupsdk.grpc_interceptors.auth import AuthInterceptor
from boulderopalscaleupsdk.grpc_interceptors.error import ErrorFormatterInterceptor
from boulderopalscaleupsdk.plotting.dtypes import Plot
from boulderopalscaleupsdk.protobuf.v1 import (
    agent_pb2,
    device_pb2,
    device_pb2_grpc,
    job_pb2,
    job_pb2_grpc,
    task_pb2,
)
from boulderopalscaleupsdk.routines import Routine
from boulderopalscaleupsdk.solutions import Solution
from google.protobuf.json_format import MessageToDict, ParseDict
from google.protobuf.struct_pb2 import Struct
from pydantic import TypeAdapter
from rich.console import Console
from rich.table import Table

from boulderopalscaleup.auth import get_default_api_key_auth
from boulderopalscaleup.constants import API_KEY_NAME, SERVER_URL
from boulderopalscaleup.controllers._base import Controller
from boulderopalscaleup.plots import Plotter

from .device import DeviceData, DeviceSummary
from .utils import display_element

if TYPE_CHECKING:
    from boulderopalscaleupsdk.common.dtypes import GrpcMetadata
    from qctrlclient import ApiKeyAuth

LOG = logging.getLogger(__name__)


def _get_package_version() -> str:
    """
    Get the version of this package.

    Returns
    -------
    str
        The package version, or "unknown" if it cannot be determined.
    """
    try:
        return metadata.version("boulder-opal-scale-up")
    except metadata.PackageNotFoundError:
        return "unknown"


class QctrlScaleUpClient:
    """
    Q-CTRL Scale Up client providing API access to experiments.
    """

    def __init__(
        self,
        controller: Controller,
        organization_slug: str,
        api_key: str | None = None,
        local_mode: bool = False,
        api_url: str = SERVER_URL,
    ):
        """
        Initialize the client.

        Parameters
        ----------
        controller : Controller
            The controller instance used to manage QPU-interfacing controllers.
        organization_slug : str
            The name of the organization using the Scale Up API.
        api_key : str or None, optional
            The API key for authenticating with the Q-CTRL server. If not provided,
            the key is retrieved from the environment variable `QCTRL_API_KEY`.
        api_url : str, optional
            The URL of the Boulder Opal Scale Up server. Defaults to the value of `SERVER_URL`.
        local_mode : bool, optional
            If True, uses a local unauthenticated server. Defaults to False.

        Raises
        ------
        RuntimeError
            If no API key is provided and the environment variable `QCTRL_API_KEY` is not set.
        """
        self.controller = controller
        self.auth: ApiKeyAuth | None
        if local_mode:
            self.auth = None
        else:
            if api_key is None:
                try:
                    api_key = os.environ[API_KEY_NAME]
                except KeyError as error:
                    raise RuntimeError(
                        "No API key provided in environment or function call. "
                        "To call this function without arguments, "
                        f"save your API key's value in the {API_KEY_NAME} "
                        "environment variable.",
                    ) from error
            self.auth = get_default_api_key_auth(api_key)
        self.agent_settings = AgentSettings(agent_id="dummy_agent_id", remote_url=api_url)
        self.request_metadata: GrpcMetadata = [
            ("organization-slug", organization_slug),
            ("client-version", _get_package_version()),
        ]

        self._current_device_name: str | None = None
        self._device_mgr = device_pb2_grpc.DeviceManagerServiceStub(self._create_channel(api_url))
        self._job_mgr = job_pb2_grpc.JobManagerServiceStub(self._create_channel(api_url))
        self._controller_resolver = ControllerResolverService()
        self._silent = False

    @property
    # Read-only to force use of set_current_device which performs checks and is async
    def current_device_name(self) -> str | None:
        return self._current_device_name

    async def create_device(self, device_name: str, device_config: Path) -> JobId:
        """
        Create and initialize a device for experiments.

        Parameters
        ----------
        device_name : str
            The name of the device to be created.
        device_config : Path
            The file path to the device configuration file.

        Raises
        ------
        ScaleUpServerError
            If the device initialization fails on the server.
        """
        device_descriptor = DeviceConfigLoader(device_config).load_device_info()
        device_data = {"device_data": device_descriptor.model_dump_json()}

        if not self._silent:
            _display_message("Creating device...")

        request = device_pb2.CreateRequest(
            device_name=device_name,
            device_data=dict_to_struct(device_data),
        )
        response: device_pb2.CreateResponse = self._device_mgr.Create(
            request,
            metadata=self.request_metadata,
        )
        if not response.done:
            raise ScaleUpServerError("Failed to create device.")

        class _Init(Routine):
            _routine_name: str = "init"

        await self.set_current_device(device_name)
        job_id = await self.run(_Init())
        if not self._silent:
            _display_message(f"Device '{device_name}' created successfully.")
        return job_id

    async def set_current_device(self, device_name: str) -> None:
        """
        Set which device to use in routines and experiments.

        Parameters
        ----------
        device_name : str
            The name of the device to mark as the current device.
        """
        if not self._silent:
            _display_message(f"Setting current device to '{device_name}'...")

        self._silent = True
        # We check device exists by retrieving it
        await self.get_device_data(device_name)
        self._silent = False

        self._current_device_name = device_name
        if not self._silent:
            _display_message(f"Current device set to '{device_name}'.")

    async def get_device_summary(self, device_name: str | None = None) -> DeviceSummary:
        """
        Retrieve the summary of a device.

        Parameters
        ----------
        device_name : str or None, optional
            The device whose summary should be retrieved. Defaults to the current device.

        Returns
        -------
        DeviceSummary
            The summary of the device.

        Raises
        ------
        RuntimeError
            If the device name is not set.
        ScaleUpServerError
            If the response from the server is invalid.
        """
        device_name = self._default_to_current_device(device_name)
        if not self._silent:
            _display_message(f"Generating device summary for device '{device_name}'...")

        request = device_pb2.GetMetadataRequest(device_name=device_name)
        response = self._device_mgr.GetMetadata(request, metadata=self.request_metadata)
        if response is None or not isinstance(response, device_pb2.GetMetadataResponse):
            raise ScaleUpServerError("Failed to retrieve device summary.")

        if not self._silent:
            _display_message(f"Device summary for '{device_name}' generated successfully.")
        return DeviceSummary.model_validate(MessageToDict(response.metadata))

    async def display_device_data_sheet(
        self,
        device_name: str | None = None,
        node_name: str | None = None,
    ) -> None:
        """
        Display a data sheet of the components in the device.

        Parameters
        ----------
        device_name : str or None, optional
            The name of the device to display. Defaults to the current device.
        node_name : str or None, optional
            The name of the node to display. If not provided, all applicable nodes are summarized.
        """
        device_name = self._default_to_current_device(device_name)
        device = await self.get_device_data(device_name)
        if node_name is None:
            display_nodes = {
                key: node
                for key, node in device.qpu.nodes.items()
                if isinstance(node, Resonator | Transmon)
            }
        else:
            node = device.qpu.nodes.get(node_name)
            if not isinstance(node, Resonator | Transmon):
                raise ValueError(f"{node_name} is not a Resonator or Transmon.")
            display_nodes = {node_name: node}
        for name, node in sorted(display_nodes.items()):
            display_element(name, node)

    async def get_device_snapshot(
        self,
        output: Path | str,
        device_name: str | None = None,
    ) -> None:
        """
        Get a snapshot of a device.

        Parameters
        ----------
        output : Path | str
            The output path to save this snapshot.
        device_name : str or None, optional
            The name of the device to snapshot. Defaults to the current device.
        """
        snapshot = await self._get_device_snapshot(device_name=device_name, version=None)
        Path(output).write_text(json.dumps(snapshot))
        snapshot_name = snapshot["device_name"]
        _display_message(f"Snapshot of device '{snapshot_name}' saved to '{output}'.")

    async def _get_device_snapshot(
        self,
        device_name: str | None = None,
        version: int | None = None,
    ) -> dict[str, Any]:
        _device_name = device_name or self._current_device_name
        if _device_name is None:
            raise RuntimeError(
                f"No current device set. Call {self.set_current_device.__name__} first.",
            )
        _display_message(f"Getting snapshot of device '{_device_name}'.")
        request = device_pb2.GetSnapshotRequest(
            device_name=_device_name,
            version=version,
        )
        resp: device_pb2.GetSnapshotResponse = self._device_mgr.GetSnapshot(
            request,
            metadata=self.request_metadata,
        )
        return MessageToDict(resp.data)

    async def set_device_snapshot(
        self,
        snapshot: Path | str,
        device_name: str | None = None,
    ) -> None:
        """
        Sets the state of a device by applying the snapshot obtained from `get_snapshot`.

        Parameters
        ----------
        snapshot : Path | str
            The path to a snapshot file.
        device_name : str or None, optional
            The name of the device to apply the snapshot. Defaults to the current device.
        """
        _display_message(
            f"Applying snapshot '{snapshot}' to device '{self.current_device_name}'.",
        )
        _snapshot_dict = json.loads(Path(snapshot).read_text())
        await self._set_device_snapshot(_snapshot_dict, device_name=device_name)
        _display_message(
            f"Snapshot '{snapshot}' applied to device '{self.current_device_name}' successfully.",
        )

    async def _set_device_snapshot(
        self,
        snapshot: dict[str, Any],
        device_name: str | None = None,
    ) -> dict[str, Any]:
        _device_name = device_name or self._current_device_name
        if _device_name is None:
            raise RuntimeError(
                f"No current device set. Call {self.set_current_device.__name__} first.",
            )
        request = device_pb2.SetSnapshotRequest(
            device_name=_device_name,
            data=dict_to_struct(snapshot),
        )
        resp: device_pb2.SetSnapshotResponse = self._device_mgr.SetSnapshot(
            request,
            metadata=self.request_metadata,
        )
        return MessageToDict(resp)

    @staticmethod
    def display(  # noqa: C901
        data: DeviceSummary | JobSummary | JobData | list[DeviceSummary] | list[JobSummary],
    ) -> None:
        console = Console()
        if isinstance(data, JobData):
            for display_item in data.get_display_items():
                if isinstance(display_item, Plot):
                    Plotter(display_item).figure.show()
                else:
                    console.print(display_item)
            return

        data_list = data if isinstance(data, list) else [data]

        if len(data_list) == 0:
            raise ValueError("Data cannot be an empty list.")

        data_type = type(data_list[0])
        if not isinstance(data_list[0], DeviceSummary | JobSummary):
            raise TypeError(
                f"Invalid data to display. Got {data_type.__name__}.",
            )

        if not all(isinstance(item, data_type) for item in data_list):
            raise TypeError(
                f"All items in the list must be of the same {data_type.__name__}.",
            )

        title = ""
        match data_list[0]:
            case JobSummary():
                title = "Job summary"
            case DeviceSummary():
                title = "Device summary"

        _data = [element.model_dump() for element in data_list]

        table = Table(title=title)
        keys = _data[0].keys()
        for key in keys:
            table.add_column(str(key))

        for item in _data:
            row = [str(item[key]) for key in keys]
            table.add_row(*row)

        console.print(table)

    async def get_device_data(
        self,
        device_name: str | None = None,
    ) -> DeviceData:
        """
        Get latest data for a device.

        Parameters
        ----------
        device_name : str or None, optional
            The name of the device. Defaults to the current device.
        """
        device_name = self._default_to_current_device(device_name)

        if not self._silent:
            _display_message(f"Retrieving data for device '{device_name}'...")

        request = device_pb2.GetDataRequest(device_name=device_name)
        response: device_pb2.GetDataResponse = self._device_mgr.GetData(
            request,
            metadata=self.request_metadata,
        )
        if (
            response.processor_data is None
            or response.controller_data is None
            or response.defcals is None
        ):
            raise ScaleUpServerError(f"Failed to retrieve {device_name} device data.")

        if not self._silent:
            _display_message(f"Data for device '{device_name}' retrieved successfully.")

        superconducting_processor = SuperconductingProcessor.from_dict(
            MessageToDict(response.processor_data),
        )
        controller_info = (
            self._controller_resolver.resolve_controller_info_from_controller_data_struct(
                response.controller_data,
            )
        )
        defcals = {}
        for item in response.defcals:
            defcal_data = DefCalData(**MessageToDict(item))
            defcals[(defcal_data.gate, tuple(defcal_data.addr))] = defcal_data

        classifier_data = ClassifierData.model_validate(
            MessageToDict(response.classifier_data),
        )
        return DeviceData(
            qpu=superconducting_processor,
            controller_info=controller_info,
            _defcals=defcals,
            _iq_classifiers=dict(classifier_data.iq_classifiers),
            _leakage_classifiers=dict(classifier_data.leakage_classifiers),
            _enabled_qubits=list(response.enabled_qubits),
        )

    async def update_device(self, new_processor_details: SuperconductingProcessor) -> None:
        """
        Update the current device's processor information.

        Parameters
        ----------
        new_processor_details : SuperconductingProcessor
            The new processor information for the current device.

        Raises
        ------
        RuntimeError
            If no current device has been set.
        """
        if self.current_device_name is None:
            raise RuntimeError(
                f"No current device set. Call {self.set_current_device.__name__} first.",
            )

        if not self._silent:
            _display_message(f"Updating device '{self.current_device_name}'...")

        request = device_pb2.UpdateRequest(
            device_name=self.current_device_name,
            processor_data=ParseDict(new_processor_details.to_dict(), Struct()),
        )
        response: device_pb2.UpdateResponse = self._device_mgr.Update(
            request,
            metadata=self.request_metadata,
        )
        if response is None or not isinstance(response, device_pb2.UpdateResponse):
            raise ScaleUpServerError("Invalid response received when updating device.")

        if not self._silent:
            _display_message(f"Device '{self.current_device_name}' updated successfully.")

    async def update_defcal(self, gate: str, addr: str | tuple[str, ...], program: str) -> JobId:
        """
        Update the defcal for a specific gate and address.

        Note that this is an experimental API that is expected
        to be changed or removed in a later release.

        Parameters
        ----------
        gate : str
            The name of the gate to update.
        addr : str | tuple[str, ...]
            The address of the qubit to update.
        program : str
            The OpenQASM program to set for the defcal.

        Returns
        -------
        JobId
            The job ID associated with the defcal update.
        """

        class _UpdateDefcal(Routine):
            _routine_name: str = "defcal_update"
            gate: str
            addr: list[str]
            program: str

        _addr = [addr] if isinstance(addr, str) else list(addr)
        return await self.run(_UpdateDefcal(gate=gate, addr=_addr, program=program))

    async def delete_device(self, device_name: str) -> None:
        """
        Delete the specified device.

        Parameters
        ----------
        device_name : str
            The name of the device to delete.
        """

        if not self._silent:
            _display_message(f"Deleting device '{device_name}'...")
        if not self._delete_device_from_server(device_name):
            raise ScaleUpServerError(f"Failed to delete device '{device_name}' from server.")

        if self.current_device_name == device_name:
            self._current_device_name = None

    async def get_devices(self) -> list[DeviceSummary]:
        """
        Retrieve a summary for all devices.

        Returns
        -------
        list[DeviceSummary]:
            The information about the devices.
        """
        if not self._silent:
            _display_message("Retrieving summary for all devices...")

        async def _get_page(next_cursor: str | None) -> tuple[list[DeviceSummary], str | None]:
            request = device_pb2.GetAllDevicesMetadataRequest(
                limit=20,
                next_cursor=next_cursor,
            )
            response = self._device_mgr.GetAllDevicesMetadata(
                request,
                metadata=self.request_metadata,
            )
            if response is None or not isinstance(
                response,
                device_pb2.GetAllDevicesMetadataResponse,
            ):
                raise ScaleUpServerError("Invalid response when attempting to get device summary.")
            summaries = [
                DeviceSummary.model_validate(MessageToDict(metadata))
                for metadata in response.metadatas
            ]
            next_cursor = (
                response.next_cursor
                if response.next_cursor is not None and len(response.next_cursor) > 0
                else None
            )
            return summaries, next_cursor

        summaries, next_cursor = await _get_page(None)
        while next_cursor is not None:
            _summaries, next_cursor = await _get_page(next_cursor)
            summaries.extend(_summaries)

        if not self._silent:
            _display_message("Summary for all devices retrieved successfully.")

        return summaries

    def _default_to_current_device(self, device_name: str | None) -> str:
        if isinstance(device_name, str):
            return device_name

        if self.current_device_name is None:
            raise RuntimeError(
                f"No current device has been set. Call {self.set_current_device.__name__} first.",
            )

        return self.current_device_name

    def _delete_device_from_server(self, device_name: str) -> bool:
        request = device_pb2.DeleteRequest(device_name=device_name)
        response = self._device_mgr.Delete(request, metadata=self.request_metadata)
        if response is None or not isinstance(response, device_pb2.DeleteResponse):
            LOG.error("Invalid response from server when attempting to delete device")
            return False
        return response.done

    def _get_channel_interceptors(self) -> list:
        """
        Get the interceptors for the gRPC channel.
        """
        interceptors: list[SyncClientInterceptor] = []
        if os.getenv("QCTRL_ENABLE_GRPC_FRIENDLY_ERRORS", "true").lower() == "true":
            interceptors.append(ErrorFormatterInterceptor(include_code=False))
        if self.auth:
            interceptors.append(AuthInterceptor(self.auth))
        return interceptors

    def _create_channel(
        self,
        api_url: str,
        interceptors: list[SyncClientInterceptor] | None = None,
    ) -> grpc.Channel:
        """
        Create a gRPC channel.
        """
        grpc_max_send_message_length = os.getenv(
            "GRPC_MAX_SEND_MESSAGE_LENGTH",
            DEFAULT_GRPC_MAX_SEND_MESSAGE_LENGTH,
        )
        grpc_max_receive_message_length = os.getenv(
            "GRPC_MAX_RECEIVE_MESSAGE_LENGTH",
            DEFAULT_GRPC_MAX_RECEIVE_MESSAGE_LENGTH,
        )
        options = [
            ("grpc.max_send_message_length", grpc_max_send_message_length),
            ("grpc.max_receive_message_length", grpc_max_receive_message_length),
        ]
        host = api_url.split(":")[0]
        force_insecure = os.getenv("QCTRL_SU_API_FORCE_INSECURE_CHANNEL", "false").lower() == "true"
        if force_insecure or host in ["localhost", "127.0.0.1", "0.0.0.0", "::"]:
            channel = grpc.insecure_channel(api_url, options=options)
        else:
            channel = grpc.secure_channel(api_url, grpc.ssl_channel_credentials(), options=options)

        if interceptors is None:
            interceptors = self._get_channel_interceptors()
        return grpc.intercept_channel(channel, *interceptors)

    async def enable_qubits(self, *elements: str) -> JobId:
        """
        Enable specified elements for running experiments.

        By default, experiments can target any element on the device. This method
        restricts targeting to only the specified elements, effectively disabling
        all others for experiment runs.

        Parameters
        ----------
        *elements : str
            The references of the transmons to enable.

        Returns
        -------
        JobId
            The job ID associated with the change.
        """
        if len(elements) == 0:
            raise ValueError("At least one element must be specified.")

        self._silent = True
        device_data = await self.get_device_data()
        self._silent = False
        for element in elements:
            node = device_data.qpu.nodes.get(element)
            if not isinstance(node, Transmon):
                raise ValueError(f"Element '{element}' is not a valid transmon.")  # noqa: TRY004

        return await self._enable(list(elements))

    async def enable_all_qubits(self) -> JobId:
        """
        Enable all qubits for running experiments.

        This method clears any previous restrictions set by `enable_qubits`,
        making all elements on the device available for experiment targeting.

        Returns
        -------
        JobId
            The job ID associated with the change.
        """
        return await self._enable("all_qubits")

    async def _enable(self, masks: list[str] | Literal["all_qubits"]) -> JobId:
        if self.current_device_name is None:
            raise RuntimeError(
                f"No current device set. Call {self.set_current_device.__name__} first.",
            )

        class _Mask(Routine):
            _routine_name: str = "mask_update"
            masks: list[str] | Literal["all_qubits"]

        return await self.run(_Mask(masks=masks))

    async def run(self, job: Experiment | Routine | Solution) -> JobId:
        """
        Execute an experiment, a routine, or a solution.

        Parameters
        ----------
        job : Experiment or Routine or Solution
            The experiment, routine, or solution to run.

        Returns
        -------
        JobId
            The job ID of the executed job.
        """
        if self.current_device_name is None:
            raise RuntimeError(
                f"No current device set. Call {self.set_current_device.__name__} first.",
            )

        self.agent = Agent(
            self.agent_settings,
            AgentTaskHandler(self),
            grpc_interceptors=self._get_channel_interceptors(),
        )

        match job:
            case Experiment():
                name = job.experiment_name
            case Routine():
                name = job.routine_name
            case Solution():
                name = job.solution_name

        job_id = await self.agent.start_session(
            metadata=self.request_metadata,
            device_name=self.current_device_name,
            routine=name,
            data=dict_to_struct(job.model_dump()),
        )

        if job_id is None:
            raise ScaleUpServerError("Failed to execute the job.")

        return job_id

    async def run_experiment(self, experiment: Experiment) -> JobId:
        """
        Execute an experiment.

        Parameters
        ----------
        experiment : Experiment
            The object containing the parameters for the experiment to be executed.

        Returns
        -------
        JobId
            The job ID of the executed experiment.

        Raises
        ------
        RuntimeError
            If the device name is not set before running the experiment.
        """

        return await self.run(experiment)

    async def run_routine(self, routine: Routine) -> JobId:
        """
        Execute a routine.

        Parameters
        ----------
        routine : Routine
            The object containing the parameters for the routine to be executed.

        Returns
        -------
        JobId
            The job ID of the executed routine.

        Raises
        ------
        RuntimeError
            If the current device name is not set before running the routine.
        """
        return await self.run(routine)

    async def run_solution(self, solution: Solution) -> JobId:
        """
        Execute a solution.

        Parameters
        ----------
        solution: Solution
            The object containing the parameters for the solution to be executed.

        Returns
        -------
        JobId
            The job ID of the executed solution.

        Raises
        ------
        RuntimeError
            If the current device name is not set before running the solution.
        """
        return await self.run(solution)

    async def get_job_data(self, job_id: str) -> JobData:
        """
        Retrieves details about a specific job executed on the device, such as
        its status, execution results, and associated metadata.

        Parameters
        ----------
        job_id : str
            The ID of the job to retrieve.

        Returns
        -------
        JobData
            The job data.

        Raises
        ------
        ScaleUpServerError
            If the response is invalid.
        """
        if not self._silent:
            _display_message(f"Retrieving data for job '{job_id}'...")

        response = self._job_mgr.Get(
            job_pb2.GetRequest(job_id=job_id),
            metadata=self.request_metadata,
        )
        if response is None:
            raise ScaleUpServerError("Invalid response.")
        if not isinstance(response, job_pb2.GetResponse):
            raise ScaleUpServerError("Unexpected response type.")

        if not self._silent:
            _display_message(f"Retrieved data for job '{job_id}'.")
        return JobData.model_validate(MessageToDict(response.job_data))

    async def get_jobs(
        self,
        device_name: str | None = None,
        job_name: str | None = None,
        page: int = DEFAULT_JOB_HISTORY_PAGE,
        limit: int = DEFAULT_JOB_HISTORY_PAGE_SIZE,
        sort_order: JobHistorySortOrder = DEFAULT_JOB_HISTORY_SORT_ORDER,
    ) -> list[JobSummary]:
        """
        Retrieves all the jobs that have been previously executed on the given device.

        Parameters
        ----------
        device_name : str
            The name of the device to filter the history by. Defaults to current.
        job_name : str, optional
            The name of the job to filter the history by. Defaults to None.
        page : int, optional
            The page number to retrieve. Defaults to 1.
        limit : int, optional
            The number of jobs to retrieve per page. Defaults to 10.
        sort_order : JobHistorySortOrder, optional
            The sort order for the results.
            Defaults to reverse chronological.

        Returns
        -------
        list[JobSummary]
            The history of jobs run.

        Raises
        ------
        ScaleUpServerError
            If the response is invalid.
        """
        device_name = self._default_to_current_device(device_name)

        if not self._silent:
            _display_message(f"Retrieving jobs for device '{device_name}'...")

        response = self._job_mgr.List(
            job_pb2.ListRequest(
                device_name=device_name,
                job_name=job_name,
                page=page,
                limit=limit,
                sort_order=sort_order.value,
            ),
            metadata=self.request_metadata,
        )
        if response is None:
            raise ScaleUpServerError("Invalid response.")
        if not isinstance(response, job_pb2.ListResponse):
            raise ScaleUpServerError("Unexpected response type.")

        if not self._silent:
            _display_message(
                f"Retrieved jobs for device '{device_name}'.",
            )
        return [JobSummary.model_validate(MessageToDict(job)) for job in response.jobs]

    async def get_job_summary(self, job_id: str) -> JobSummary:
        """
        Retrieves a summary of a specific job executed on the device.

        Parameters
        ----------
        job_id : str
            The ID of the job to retrieve.

        Returns
        -------
        JobSummary
            The job summary.

        Raises
        ------
        ScaleUpServerError
            If the response is invalid.
        """
        if not self._silent:
            _display_message(f"Retrieving summary for job '{job_id}'...")

        response = self._job_mgr.GetSummary(
            job_pb2.GetSummaryRequest(job_id=job_id),
            metadata=self.request_metadata,
        )
        if response is None:
            raise ScaleUpServerError("Invalid response.")
        if not isinstance(response, job_pb2.GetSummaryResponse):
            raise ScaleUpServerError("Unexpected response type.")

        if not self._silent:
            _display_message(f"Retrieved summary for job '{job_id}'.")
        return JobSummary.model_validate(MessageToDict(response.job_summary_data))


class AgentTaskHandler(TaskHandler):
    def __init__(self, client: QctrlScaleUpClient) -> None:
        self._client = client

    async def handle(
        self,
        request: agent_pb2.RunProgramRequest
        | agent_pb2.RunQuantumMachinesMixerCalibrationRequest
        | agent_pb2.DisplayResultsRequest
        | agent_pb2.AskRequest,
    ) -> (
        agent_pb2.RunProgramResponse
        | agent_pb2.RunQuantumMachinesMixerCalibrationResponse
        | agent_pb2.DisplayResultsResponse
        | agent_pb2.AskResponse
        | task_pb2.TaskErrorDetail
    ):
        match request:
            case agent_pb2.RunProgramRequest():
                return await self._client.controller.run_program(request)
            case agent_pb2.RunQuantumMachinesMixerCalibrationRequest():
                return await self._client.controller.run_mixer_calibration(request)
            case agent_pb2.DisplayResultsRequest():
                return await _display_results(request)
            case agent_pb2.AskRequest():
                return await _ask_user(request)


@dataclass
class CalibrationError:
    message: str


async def _display_results(
    results: agent_pb2.DisplayResultsRequest,
) -> agent_pb2.DisplayResultsResponse:
    """
    Display results to the user.
    """

    LOG.info("Displaying results")

    if results.plots is not None:
        for plot in results.plots:
            plot_data: Plot = TypeAdapter(Plot).validate_json(plot)
            Plotter(plot_data).figure.show()

    if results.message is not None:
        _display_message(results.message)

    return agent_pb2.DisplayResultsResponse()


async def _handle_yes_no_ask(ask: agent_pb2.AskRequest) -> agent_pb2.AskResponse:
    """
    Handle yes/no questions from the user.
    """
    while True:
        response = input(f"{ask.message}: ").strip().lower()
        if response in ("y", "yes"):
            return agent_pb2.AskResponse(response="y")
        if response in ("n", "no"):
            return agent_pb2.AskResponse(response="n")
        _display_message("Please answer with 'y' or 'n'.")


async def _ask_user(ask: agent_pb2.AskRequest) -> agent_pb2.AskResponse:
    """
    Ask user interactively.
    """
    _display_message(ask.message)
    match ask.expected_response_type:
        case "yes_no":
            return await _handle_yes_no_ask(ask)
        case _:
            raise ScaleUpServerError(
                f"Unknown expected response type '{ask.expected_response_type}'",
            )


def dict_to_struct(dictionary: dict) -> Struct:
    result = Struct()
    result.update(dictionary)
    return result


def _display_message(message: str) -> None:
    console = Console()
    console.is_jupyter = False
    console.print(message)
